"""
This module contains enumerations used in pyproj.crs.
"""
from pyproj.enums import BaseEnum


class DatumType(BaseEnum):
    """
    .. versionadded:: 2.5.0

    Datum Types for creating datum with :meth:`pyproj.crs.Datum.from_name`

    Attributes
    ----------
    GEODETIC_REFERENCE_FRAME
    DYNAMIC_GEODETIC_REFERENCE_FRAME
    VERTICAL_REFERENCE_FRAME
    DYNAMIC_VERTICAL_REFERENCE_FRAME
    DATUM_ENSEMBLE
    """

    GEODETIC_REFERENCE_FRAME = "GEODETIC_REFERENCE_FRAME"
    DYNAMIC_GEODETIC_REFERENCE_FRAME = "DYNAMIC_GEODETIC_REFERENCE_FRAME"
    VERTICAL_REFERENCE_FRAME = "VERTICAL_REFERENCE_FRAME"
    DYNAMIC_VERTICAL_REFERENCE_FRAME = "DYNAMIC_VERTICAL_REFERENCE_FRAME"
    DATUM_ENSEMBLE = "DATUM_ENSEMBLE"


class CoordinateOperationType(BaseEnum):
    """
    .. versionadded:: 2.5.0

    Coordinate Operation Types for creating operation
    with :meth:`pyproj.crs.CoordinateOperation.from_name`

    Attributes
    ----------
    CONVERSION
    TRANSFORMATION
    CONCATENATED_OPERATION
    OTHER_COORDINATE_OPERATION
    """

    CONVERSION = "CONVERSION"
    TRANSFORMATION = "TRANSFORMATION"
    CONCATENATED_OPERATION = "CONCATENATED_OPERATION"
    OTHER_COORDINATE_OPERATION = "OTHER_COORDINATE_OPERATION"


class Cartesian2DCSAxis(BaseEnum):
    """
    .. versionadded:: 2.5.0

    Cartesian 2D Coordinate System Axis for creating axis with
    with :class:`pyproj.crs.coordinate_system.Cartesian2DCS`

    Attributes
    ----------
    EASTING_NORTHING
    NORTHING_EASTING
    EASTING_NORTHING_FT
    NORTHING_EASTING_FT
    EASTING_NORTHING_US_FT
    NORTHING_EASTING_US_FT
    NORTH_POLE_EASTING_SOUTH_NORTHING_SOUTH
    SOUTH_POLE_EASTING_NORTH_NORTHING_NORTH
    WESTING_SOUTHING
    """

    EASTING_NORTHING = "EASTING_NORTHING"
    NORTHING_EASTING = "NORTHING_EASTING"
    EASTING_NORTHING_FT = "EASTING_NORTHING_FT"
    NORTHING_EASTING_FT = "NORTHING_EASTING_FT"
    EASTING_NORTHING_US_FT = "EASTING_NORTHING_US_FT"
    NORTHING_EASTING_US_FT = "NORTHING_EASTING_US_FT"
    NORTH_POLE_EASTING_SOUTH_NORTHING_SOUTH = "NORTH_POLE_EASTING_SOUTH_NORTHING_SOUTH"
    SOUTH_POLE_EASTING_NORTH_NORTHING_NORTH = "SOUTH_POLE_EASTING_NORTH_NORTHING_NORTH"
    WESTING_SOUTHING = "WESTING_SOUTHING"


class Ellipsoidal2DCSAxis(BaseEnum):
    """
    .. versionadded:: 2.5.0

    Ellisoidal 2D Coordinate System Axis for creating axis with
    with :class:`pyproj.crs.coordinate_system.Ellipsoidal2DCS`

    Attributes
    ----------
    LONGITUDE_LATITUDE
    LATITUDE_LONGITUDE
    """

    LONGITUDE_LATITUDE = "LONGITUDE_LATITUDE"
    LATITUDE_LONGITUDE = "LATITUDE_LONGITUDE"


class Ellipsoidal3DCSAxis(BaseEnum):
    """
    .. versionadded:: 2.5.0

    Ellisoidal 3D Coordinate System Axis for creating axis with
    with :class:`pyproj.crs.coordinate_system.Ellipsoidal3DCS`

    Attributes
    ----------
    LONGITUDE_LATITUDE_HEIGHT
    LATITUDE_LONGITUDE_HEIGHT
    """

    LONGITUDE_LATITUDE_HEIGHT = "LONGITUDE_LATITUDE_HEIGHT"
    LATITUDE_LONGITUDE_HEIGHT = "LATITUDE_LONGITUDE_HEIGHT"


class VerticalCSAxis(BaseEnum):
    """
    .. versionadded:: 2.5.0

    Vertical Coordinate System Axis for creating axis with
    with :class:`pyproj.crs.coordinate_system.VerticalCS`

    Attributes
    ----------
    UP
    UP_FT
    UP_US_FT
    DEPTH
    DEPTH_FT
    DEPTH_US_FT
    GRAVITY_HEIGHT
    GRAVITY_HEIGHT_FT
    GRAVITY_HEIGHT_US_FT
    """

    GRAVITY_HEIGHT = "GRAVITY_HEIGHT"
    GRAVITY_HEIGHT_FT = "GRAVITY_HEIGHT_FT"
    GRAVITY_HEIGHT_US_FT = "GRAVITY_HEIGHT_US_FT"
    DEPTH = "DEPTH"
    DEPTH_FT = "DEPTH_FT"
    DEPTH_US_FT = "DEPTH_US_FT"
    UP = "UP"
    UP_FT = "UP_FT"
    UP_US_FT = "UP_US_FT"



import pytest
from numpy.testing import assert_almost_equal

from pyproj import (
    get_angular_units_map,
    get_authorities,
    get_codes,
    get_ellps_map,
    get_prime_meridians_map,
    get_proj_operations_map,
    get_units_map,
)
from pyproj._list import Unit
from pyproj.enums import PJType


def test_units_map():
    with pytest.warns(DeprecationWarning):
        units_map = get_units_map()
    assert isinstance(units_map["m"], Unit)
    assert units_map["m"].id == "m"
    assert units_map["m"].name == "Meter"


def test_angular_units_map():
    with pytest.warns(DeprecationWarning):
        ang_map = get_angular_units_map()
    assert isinstance(ang_map["deg"], Unit)
    assert ang_map["deg"].id == "deg"
    assert ang_map["deg"].name == "Degree"


def test_get_ellps_map():
    ellps_map = get_ellps_map()
    assert ellps_map["WGS84"]["description"] == "WGS 84"
    assert_almost_equal(ellps_map["WGS84"]["a"], 6378137.0, decimal=1)
    assert_almost_equal(ellps_map["WGS84"]["rf"], 298.257223563, decimal=1)


def test_get_prime_meridians_map():
    prime_meridians_map = get_prime_meridians_map()
    assert prime_meridians_map["greenwich"] == "0dE"


def test_get_proj_operations_map():
    proj_operations_map = get_proj_operations_map()
    assert proj_operations_map["aea"] == "Albers Equal Area"


def test_get_authorities():
    assert "EPSG" in get_authorities()


@pytest.mark.parametrize(
    "auth, pj_type, deprecated",
    [
        ("IGNF", PJType.ELLIPSOID, False),
        ("EPSG", PJType.CRS, False),
        ("EPSG", PJType.CRS, True),
        ("PROJ", PJType.ELLIPSOID, False),
        ("IGNF", "ELLIPSOID", False),
        ("EPSG", "CRS", False),
        ("EPSG", "crs", True),
        ("PROJ", "ellipsoid", False),
    ],
)
def test_get_codes(auth, pj_type, deprecated):
    assert get_codes(auth, pj_type, deprecated)


@pytest.mark.parametrize(
    "auth, pj_type",
    [("blob", "BOUND_CRS"), ("PROJ", PJType.BOUND_CRS), ("ITRF", PJType.BOUND_CRS)],
)
def test_get_codes__empty(auth, pj_type):
    assert not get_codes(auth, pj_type)


def test_get_codes__invalid_auth():
    with pytest.raises(TypeError):
        get_codes(123, PJType.BOUND_CRS)


def test_get_codes__invalid_code():
    with pytest.raises(ValueError):
        get_codes("ITRF", "frank")



from numpy.testing import assert_almost_equal

from pyproj.crs.datum import (
    CustomDatum,
    CustomEllipsoid,
    CustomPrimeMeridian,
    Ellipsoid,
    PrimeMeridian,
)


def test_custom_datum():
    cd = CustomDatum()
    assert cd.ellipsoid.name == "WGS 84"
    assert cd.prime_meridian.name == "Greenwich"


def test_custom_datum__input():
    cd = CustomDatum(
        ellipsoid=Ellipsoid.from_epsg(7001),
        prime_meridian=PrimeMeridian.from_name("Lisbon"),
    )
    assert cd.ellipsoid.name == "Airy 1830"
    assert cd.prime_meridian.name == "Lisbon"


def test_custom_ellipsoid():
    ce = CustomEllipsoid(semi_major_axis=6378137, inverse_flattening=298.257222101)
    assert ce.name == "undefined"
    assert ce.semi_major_metre == 6378137
    assert ce.semi_minor_metre == 6356752.314140356
    assert_almost_equal(ce.inverse_flattening, 298.257222101)
    assert sorted(ce.to_json_dict()) == [
        "$schema",
        "inverse_flattening",
        "name",
        "semi_major_axis",
        "type",
    ]


def test_custom_ellipsoid__minor():
    ce = CustomEllipsoid(
        name="test", semi_major_axis=6378137, semi_minor_axis=6356752.314
    )
    assert ce.name == "test"
    assert ce.semi_major_metre == 6378137
    assert ce.semi_minor_metre == 6356752.314
    assert_almost_equal(ce.inverse_flattening, 298.25722014)
    assert sorted(ce.to_json_dict()) == [
        "$schema",
        "name",
        "semi_major_axis",
        "semi_minor_axis",
        "type",
    ]


def test_custom_ellipsoid__radius():
    ce = CustomEllipsoid(radius=6378137)
    assert ce.name == "undefined"
    assert ce.semi_major_metre == 6378137
    assert ce.semi_minor_metre == 6378137
    assert ce.inverse_flattening == 0
    assert sorted(ce.to_json_dict()) == ["$schema", "name", "radius", "type"]


def test_custom_prime_meridian():
    pm = CustomPrimeMeridian(longitude=2)
    assert pm.name == "undefined"
    assert pm.longitude == 2


def test_custom_prime_meridian__name():
    pm = CustomPrimeMeridian(longitude=1, name="frank")
    assert pm.name == "frank"
    assert pm.longitude == 1



import os
import shutil
import tempfile
from contextlib import contextmanager

import pytest
from mock import patch

import pyproj
from pyproj import CRS
from pyproj._datadir import pyproj_global_context_initialize
from pyproj.datadir import DataDirError, append_data_dir, get_data_dir, set_data_dir


def unset_data_dir():
    pyproj.datadir._USER_PROJ_DATA = None
    pyproj.datadir._VALIDATED_PROJ_DATA = None


def create_projdb(tmpdir):
    with open(os.path.join(tmpdir, "proj.db"), "w") as pjdb:
        pjdb.write("DUMMY proj.db")


@contextmanager
def proj_env():
    """
    Ensure environment variable the same at the end of the test.
    """
    unset_data_dir()
    try:
        yield
    finally:
        # make sure the data dir is cleared
        unset_data_dir()


@contextmanager
def temporary_directory():
    """
    Get a temporary directory
    """
    temp_dir = tempfile.mkdtemp()
    try:
        yield temp_dir
    finally:
        shutil.rmtree(temp_dir)


_INVALID_PATH = "/invalid/path/to/nowhere"


def setup_os_mock(os_mock, abspath_return=_INVALID_PATH, proj_dir=None):
    os_mock.path.abspath.return_value = abspath_return
    os_mock.path.join = os.path.join
    os_mock.path.dirname = os.path.dirname
    os_mock.path.exists = os.path.exists
    os_mock.pathsep = os.pathsep
    if proj_dir is None:
        os_mock.environ = {}
    else:
        os_mock.environ = {"PROJ_LIB": proj_dir}


def test_get_data_dir__missing():
    with proj_env(), pytest.raises(DataDirError), patch(
        "pyproj.datadir.find_executable", return_value=None
    ), patch("pyproj.datadir.os") as os_mock, patch("pyproj.datadir.sys") as sys_mock:
        sys_mock.prefix = _INVALID_PATH
        setup_os_mock(os_mock)
        assert get_data_dir() is None


def test_pyproj_global_context_initialize__datadir_missing():
    with proj_env(), pytest.raises(DataDirError), patch(
        "pyproj._datadir.get_data_dir", side_effect=DataDirError("test")
    ):
        pyproj_global_context_initialize()


def test_get_data_dir__from_user():
    with proj_env(), temporary_directory() as tmpdir, patch(
        "pyproj.datadir.os"
    ) as os_mock, patch(
        "pyproj.datadir.sys"
    ) as sys_mock, temporary_directory() as tmpdir_env:  # noqa: E501
        setup_os_mock(
            os_mock,
            abspath_return=os.path.join(tmpdir, "randomfilename.py"),
            proj_dir=tmpdir_env,
        )
        sys_mock.prefix = tmpdir_env
        create_projdb(tmpdir)
        create_projdb(tmpdir_env)
        set_data_dir(tmpdir)
        internal_proj_dir = os.path.join(tmpdir, "proj_dir", "share", "proj")
        os.makedirs(internal_proj_dir)
        create_projdb(internal_proj_dir)
        assert get_data_dir() == tmpdir


def test_get_data_dir__internal():
    with proj_env(), temporary_directory() as tmpdir, patch(
        "pyproj.datadir.os"
    ) as os_mock, temporary_directory() as tmpdir_fake, patch(
        "pyproj.datadir.sys"
    ) as sys_mock:
        setup_os_mock(
            os_mock,
            abspath_return=os.path.join(tmpdir, "randomfilename.py"),
            proj_dir=tmpdir_fake,
        )
        sys_mock.prefix = tmpdir_fake
        create_projdb(tmpdir)
        create_projdb(tmpdir_fake)
        internal_proj_dir = os.path.join(tmpdir, "proj_dir", "share", "proj")
        os.makedirs(internal_proj_dir)
        create_projdb(internal_proj_dir)
        assert get_data_dir() == internal_proj_dir


def test_get_data_dir__from_env_var():
    with proj_env(), temporary_directory() as tmpdir, patch(
        "pyproj.datadir.os"
    ) as os_mock, patch("pyproj.datadir.sys") as sys_mock:
        setup_os_mock(os_mock, proj_dir=tmpdir)
        sys_mock.prefix = _INVALID_PATH
        create_projdb(tmpdir)
        assert get_data_dir() == tmpdir


def test_get_data_dir__from_env_var__multiple():
    with proj_env(), temporary_directory() as tmpdir, patch(
        "pyproj.datadir.os"
    ) as os_mock, patch("pyproj.datadir.sys") as sys_mock:
        setup_os_mock(os_mock, proj_dir=os.pathsep.join([tmpdir, tmpdir, tmpdir]))
        sys_mock.prefix = _INVALID_PATH
        create_projdb(tmpdir)
        assert get_data_dir() == os.pathsep.join([tmpdir, tmpdir, tmpdir])


def test_get_data_dir__from_prefix():
    with proj_env(), temporary_directory() as tmpdir, patch(
        "pyproj.datadir.os"
    ) as os_mock, patch("pyproj.datadir.sys") as sys_mock:
        setup_os_mock(os_mock)
        sys_mock.prefix = tmpdir
        proj_dir = os.path.join(tmpdir, "share", "proj")
        os.makedirs(proj_dir)
        create_projdb(proj_dir)
        assert get_data_dir() == proj_dir


def test_get_data_dir__from_path():
    with proj_env(), temporary_directory() as tmpdir, patch(
        "pyproj.datadir.os"
    ) as os_mock, patch("pyproj.datadir.sys") as sys_mock, patch(
        "pyproj.datadir.find_executable"
    ) as find_exe:
        setup_os_mock(os_mock)
        sys_mock.prefix = _INVALID_PATH
        find_exe.return_value = os.path.join(tmpdir, "bin", "proj")
        proj_dir = os.path.join(tmpdir, "share", "proj")
        os.makedirs(proj_dir)
        create_projdb(proj_dir)
        assert get_data_dir() == proj_dir


def test_append_data_dir__internal():
    with proj_env(), temporary_directory() as tmpdir, patch(
        "pyproj.datadir.os"
    ) as os_mock:
        setup_os_mock(os_mock, os.path.join(tmpdir, "randomfilename.py"))
        create_projdb(tmpdir)
        internal_proj_dir = os.path.join(tmpdir, "proj_dir", "share", "proj")
        os.makedirs(internal_proj_dir)
        create_projdb(internal_proj_dir)
        extra_datadir = str(os.path.join(tmpdir, "extra_datumgrids"))
        append_data_dir(extra_datadir)
        assert get_data_dir() == os.pathsep.join([internal_proj_dir, extra_datadir])


def test_creating_multiple_crs_without_file_limit():
    assert [CRS.from_epsg(4326) for _ in range(1200)]
